@@ -33,6 +33,7 @@ PGrnSequentialSearchDataInitialize(PGrnSequentialSearchData *data)
 										  NULL,
 										  GRN_OBJ_COLUMN_VECTOR,
 										  grn_ctx_at(ctx, GRN_DB_TEXT));
+	data->targetColumn = NULL;
 	data->recordID = grn_table_add(ctx,
 								   data->table,
 								   NULL, 0,
@@ -72,21 +73,8 @@ PGrnSequentialSearchDataFinalize(PGrnSequentialSearchData *data)
 }
 
 static void
-PGrnSequentialSearchDataExecutePrepareIndex(PGrnSequentialSearchData *data,
-											Oid indexOID,
-											grn_obj *source)
+PGrnSequentialSearchDataExecuteClearIndex(PGrnSequentialSearchData *data)
 {
-	if (data->indexOID == indexOID)
-	{
-		if (data->indexColumnSource != source)
-		{
-			if (data->indexColumn)
-				PGrnIndexColumnClearSources(InvalidRelation, data->indexColumn);
-			data->indexColumnSource = NULL;
-		}
-		return;
-	}
-
 	if (data->indexColumn)
 	{
 		grn_obj_remove(ctx, data->indexColumn);
@@ -107,68 +95,68 @@ PGrnSequentialSearchDataExecutePrepareIndex(PGrnSequentialSearchData *data,
 		data->expression = NULL;
 	}
 
-	data->indexColumnSource = NULL;
 	data->useIndex = false;
 	data->exprFlags = PGRN_EXPR_QUERY_PARSE_FLAGS;
 }
 
-static void
-PGrnSequentialSearchDataExecuteSetIndex(PGrnSequentialSearchData *data,
-										Oid indexOID,
-										grn_obj *source,
-										const char *indexName,
-										unsigned int indexNameSize)
+static bool
+PGrnSequentialSearchDataExecutePrepareIndex(PGrnSequentialSearchData *data,
+											const char *indexName,
+											unsigned int indexNameSize,
+											PGrnSequentialSearchType type)
 {
-	grn_column_flags indexFlags =
-		GRN_OBJ_COLUMN_INDEX |
-		GRN_OBJ_WITH_POSITION;
+	const char *tag = "[sequential-search][index]";
+	Oid indexOID = InvalidOid;
+	grn_column_flags indexFlags = GRN_OBJ_COLUMN_INDEX;
 
+	if (indexNameSize > 0)
+	{
+		grn_obj *text = &(buffers->general);
+
+		grn_obj_reinit(ctx, text, GRN_DB_TEXT, 0);
+		GRN_TEXT_SET(ctx, text, indexName, indexNameSize);
+		GRN_TEXT_PUTC(ctx, text, '\0');
+		indexOID = PGrnPGIndexNameToID(GRN_TEXT_VALUE(text));
+	}
+
+	if (data->indexOID == indexOID &&
+		data->indexColumnSource == data->targetColumn &&
+		data->type == type)
+	{
+		return false;
+	}
+
+	PGrnSequentialSearchDataExecuteClearIndex(data);
 	data->useIndex = (PGrnIsTemporaryIndexSearchAvailable &&
 					  (OidIsValid(indexOID) ||
-					   source == data->textsColumn));
+					   data->targetColumn == data->textsColumn));
 
-	if (grn_obj_is_vector_column(ctx, source))
+	if (grn_obj_is_vector_column(ctx, data->targetColumn))
 		indexFlags |= GRN_OBJ_WITH_SECTION;
 
 	if (data->indexOID != indexOID)
 	{
 		Relation index;
 		bool isPGroongaIndex;
-		grn_obj *tokenizer = NULL;
-		grn_obj *normalizers = NULL;
-		grn_obj *tokenFilters = NULL;
-		grn_table_flags tableFlags = 0;
 
 		index = PGrnPGResolveIndexID(indexOID);
 		isPGroongaIndex = PGrnIndexIsPGroonga(index);
 		if (!isPGroongaIndex) {
 			RelationClose(index);
-			ereport(ERROR,
-					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-					 errmsg("pgroonga: [sequential-search][index][invalid] "
-							"not PGroonga index: <%.*s>",
-							indexNameSize,
-							indexName)));
+			PGrnCheckRC(GRN_INVALID_ARGUMENT,
+						"%s[invalid] not PGroonga index: <%.*s>",
+						tag,
+						indexNameSize, indexName);
 		}
 
-		PGrnApplyOptionValues(index,
-							  -1,
-							  PGRN_OPTION_USE_CASE_FULL_TEXT_SEARCH,
-							  &tokenizer, PGRN_DEFAULT_TOKENIZER,
-							  &normalizers, PGRN_DEFAULT_NORMALIZERS,
-							  &tokenFilters,
-							  &tableFlags,
-							  &indexFlags);
+		data->lexicon = PGrnCreateSimilarTemporaryLexicon(index);
+
 		data->exprFlags |= PGrnOptionsGetExprParseFlags(index);
+
 		RelationClose(index);
 
-		data->lexicon = PGrnCreateTable(InvalidRelation,
-										NULL,
-										tableFlags,
-										grn_ctx_at(ctx, GRN_DB_SHORT_TEXT),
-										tokenizer,
-										normalizers,
-										tokenFilters);
+		if (grn_obj_get_info(ctx, data->lexicon, GRN_INFO_DEFAULT_TOKENIZER, NULL))
+			indexFlags |= GRN_OBJ_WITH_POSITION;
 		data->indexColumn =
 			PGrnCreateColumn(InvalidRelation,
 							 data->lexicon,
@@ -180,11 +168,20 @@ PGrnSequentialSearchDataExecuteSetIndex(PGrnSequentialSearchData *data,
 
 	if (!data->lexicon)
 	{
-		grn_obj *tokenizer;
+		grn_obj *tokenizer = NULL;
 		grn_obj *normalizers = &(buffers->normalizers);
 		grn_table_flags tableFlags = GRN_OBJ_TABLE_PAT_KEY;
 
-		tokenizer = PGrnLookup(PGRN_DEFAULT_TOKENIZER, ERROR);
+		switch (type)
+		{
+		case PGRN_SEQUENTIAL_SEARCH_EQUAL_QUERY:
+			tokenizer = NULL;
+			break;
+		default:
+			tokenizer = PGrnLookup(PGRN_DEFAULT_TOKENIZER, ERROR);
+			indexFlags |= GRN_OBJ_WITH_POSITION;
+			break;
+		}
 		GRN_TEXT_SETS(ctx, normalizers, PGRN_DEFAULT_NORMALIZERS);
 		data->lexicon = PGrnCreateTable(InvalidRelation,
 										NULL,
@@ -201,81 +198,50 @@ PGrnSequentialSearchDataExecuteSetIndex(PGrnSequentialSearchData *data,
 							 data->table);
 	}
 
-	if (data->indexColumnSource != source)
+	if (data->useIndex)
 	{
-		if (data->useIndex)
-		{
-			PGrnIndexColumnSetSource(InvalidRelation,
-									 data->indexColumn,
-									 source);
-		}
-		else
-		{
-			PGrnIndexColumnClearSources(InvalidRelation, data->indexColumn);
-		}
-		data->indexColumnSource = source;
+		PGrnIndexColumnSetSource(InvalidRelation,
+								 data->indexColumn,
+								 data->targetColumn);
 	}
+	else
+	{
+		PGrnIndexColumnClearSources(InvalidRelation, data->indexColumn);
+	}
+	data->indexColumnSource = data->targetColumn;
+
+	return true;
 }
 
 static void
 PGrnSequentialSearchDataPrepare(PGrnSequentialSearchData *data,
 								grn_obj *column,
-								grn_obj *value,
-								const char *indexName,
-								unsigned int indexNameSize)
+								grn_obj *value)
 {
-	Oid indexOID = InvalidOid;
-
-	if (indexNameSize > 0)
-	{
-		grn_obj *text = &(buffers->general);
-
-		grn_obj_reinit(ctx, text, GRN_DB_TEXT, 0);
-		GRN_TEXT_SET(ctx, text, indexName, indexNameSize);
-		GRN_TEXT_PUTC(ctx, text, '\0');
-		indexOID = PGrnPGIndexNameToID(GRN_TEXT_VALUE(text));
-	}
-	PGrnSequentialSearchDataExecutePrepareIndex(data, indexOID, column);
-
 	grn_obj_set_value(ctx,
 					  column,
 					  data->recordID,
 					  value,
 					  GRN_OBJ_SET);
-
-	if (PGrnIsTemporaryIndexSearchAvailable)
-	{
-		PGrnSequentialSearchDataExecuteSetIndex(data,
-												indexOID,
-												column,
-												indexName,
-												indexNameSize);
-	}
+	data->targetColumn = column;
 }
 
 void
 PGrnSequentialSearchDataPrepareText(PGrnSequentialSearchData *data,
 									const char *target,
-									unsigned int targetSize,
-									const char *indexName,
-									unsigned int indexNameSize)
+									unsigned int targetSize)
 {
 	grn_obj *text = &(buffers->text);
-
 	GRN_TEXT_SET(ctx, text, target, targetSize);
 	PGrnSequentialSearchDataPrepare(data,
 									data->textColumn,
-									text,
-									indexName,
-									indexNameSize);
+									text);
 }
 
 void
 PGrnSequentialSearchDataPrepareTexts(PGrnSequentialSearchData *data,
 									 ArrayType *targets,
-									 grn_obj *isTargets,
-									 const char *indexName,
-									 unsigned int indexNameSize)
+									 grn_obj *isTargets)
 {
 	grn_obj *texts = &(buffers->texts);
 	ArrayIterator iterator;
@@ -319,19 +285,26 @@ PGrnSequentialSearchDataPrepareTexts(PGrnSequentialSearchData *data,
 
 	PGrnSequentialSearchDataPrepare(data,
 									data->textsColumn,
-									texts,
-									indexName,
-									indexNameSize);
+									texts);
 }
 
 static bool
 PGrnSequentialSearchDataPrepareExpression(PGrnSequentialSearchData *data,
 										  const char *expressionData,
 										  unsigned int expressionDataSize,
+										  const char *indexName,
+										  unsigned int indexNameSize,
 										  PGrnSequentialSearchType type)
 {
+	const char *tag = "[sequential-search][expression]";
+	bool indexUpdated;
 	uint64_t expressionHash;
 
+	indexUpdated = PGrnSequentialSearchDataExecutePrepareIndex(data,
+															   indexName,
+															   indexNameSize,
+															   type);
+
 	if (data->type != type)
 	{
 		data->expressionHash = 0;
@@ -339,7 +312,7 @@ PGrnSequentialSearchDataPrepareExpression(PGrnSequentialSearchData *data,
 	}
 
 	expressionHash = XXH64(expressionData, expressionDataSize, 0);
-	if (data->expressionHash == expressionHash)
+	if (!indexUpdated && data->expressionHash == expressionHash)
 		return true;
 
 	if (data->expression)
@@ -354,10 +327,9 @@ PGrnSequentialSearchDataPrepareExpression(PGrnSequentialSearchData *data,
 							  data->variable);
 	if (!data->expression)
 	{
-		ereport(ERROR,
-				(errcode(ERRCODE_OUT_OF_MEMORY),
-				 errmsg("pgroonga: failed to create expression: %s",
-						ctx->errbuf)));
+		PGrnCheckRC(GRN_NO_MEMORY_AVAILABLE,
+					"%s failed to create expression",
+					tag);
 	}
 
 	data->expressionHash = expressionHash;
@@ -368,11 +340,17 @@ PGrnSequentialSearchDataPrepareExpression(PGrnSequentialSearchData *data,
 void
 PGrnSequentialSearchDataSetMatchTerm(PGrnSequentialSearchData *data,
 									 const char *term,
-									 unsigned int termSize)
+									 unsigned int termSize,
+									 const char *indexName,
+									 unsigned int indexNameSize)
 {
+	const char *tag = "[sequential-search][match-term]";
+
 	if (PGrnSequentialSearchDataPrepareExpression(data,
 												  term,
 												  termSize,
+												  indexName,
+												  indexNameSize,
 												  PGRN_SEQUENTIAL_SEARCH_MATCH_TERM))
 	{
 		return;
@@ -380,97 +358,168 @@ PGrnSequentialSearchDataSetMatchTerm(PGrnSequentialSearchData *data,
 
 	grn_expr_append_obj(ctx,
 						data->expression,
-						data->indexColumnSource,
+						data->targetColumn,
 						GRN_OP_GET_VALUE,
 						1);
-	PGrnCheck("match-term: append match target column");
+	PGrnCheck("%s append match target column", tag);
 	grn_expr_append_const_str(ctx,
 							  data->expression,
 							  term,
 							  termSize,
 							  GRN_OP_PUSH,
 							  1);
-	PGrnCheck("match-term: append term to be matched");
+	PGrnCheck("%s append term to be matched", tag);
 	grn_expr_append_op(ctx,
 					   data->expression,
 					   GRN_OP_MATCH,
 					   2);
-	PGrnCheck("match-term: append match operator");
+	PGrnCheck("%s append match operator", tag);
+}
+
+void
+PGrnSequentialSearchDataSetEqualText(PGrnSequentialSearchData *data,
+									 const char *other,
+									 unsigned int otherSize,
+									 const char *indexName,
+									 unsigned int indexNameSize)
+{
+	const char *tag = "[sequential-search][equal-text]";
+
+	if (PGrnSequentialSearchDataPrepareExpression(data,
+												  other,
+												  otherSize,
+												  indexName,
+												  indexNameSize,
+												  PGRN_SEQUENTIAL_SEARCH_EQUAL_TEXT))
+	{
+		return;
+	}
+
+	grn_expr_append_obj(ctx,
+						data->expression,
+						data->targetColumn,
+						GRN_OP_GET_VALUE,
+						1);
+	PGrnCheck("%s append match target column", tag);
+	grn_expr_append_const_str(ctx,
+							  data->expression,
+							  other,
+							  otherSize,
+							  GRN_OP_PUSH,
+							  1);
+	PGrnCheck("%s append equal text", tag);
+	grn_expr_append_op(ctx,
+					   data->expression,
+					   GRN_OP_EQUAL,
+					   2);
+	PGrnCheck("%s append equal operator", tag);
+}
+
+void
+PGrnSequentialSearchDataSetPrefix(PGrnSequentialSearchData *data,
+								  const char *prefix,
+								  unsigned int prefixSize,
+								  const char *indexName,
+								  unsigned int indexNameSize)
+{
+	const char *tag = "[sequential-search][prefix]";
+
+	if (PGrnSequentialSearchDataPrepareExpression(data,
+												  prefix,
+												  prefixSize,
+												  indexName,
+												  indexNameSize,
+												  PGRN_SEQUENTIAL_SEARCH_PREFIX))
+	{
+		return;
+	}
+
+	grn_expr_append_obj(ctx,
+						data->expression,
+						data->targetColumn,
+						GRN_OP_GET_VALUE,
+						1);
+	PGrnCheck("%s append match target column", tag);
+	grn_expr_append_const_str(ctx,
+							  data->expression,
+							  prefix,
+							  prefixSize,
+							  GRN_OP_PUSH,
+							  1);
+	PGrnCheck("%s append prefix", tag);
+	grn_expr_append_op(ctx,
+					   data->expression,
+					   GRN_OP_PREFIX,
+					   2);
+	PGrnCheck("%s append prefix operator", tag);
 }
 
 void
 PGrnSequentialSearchDataSetQuery(PGrnSequentialSearchData *data,
 								 const char *query,
-								 unsigned int querySize)
+								 unsigned int querySize,
+								 const char *indexName,
+								 unsigned int indexNameSize,
+								 PGrnSequentialSearchType type)
 {
-	grn_rc rc;
+	const char *tag = "[sequential-search][query]";
 
 	if (PGrnSequentialSearchDataPrepareExpression(data,
 												  query,
 												  querySize,
-												  PGRN_SEQUENTIAL_SEARCH_QUERY))
+												  indexName,
+												  indexNameSize,
+												  type))
 	{
 		return;
 	}
 
-	rc = grn_expr_parse(ctx,
-						data->expression,
-						query, querySize,
-						data->indexColumnSource,
-						GRN_OP_MATCH, GRN_OP_AND,
-						data->exprFlags);
-	if (rc != GRN_SUCCESS)
-	{
-		char message[GRN_CTX_MSGSIZE];
-		grn_strncpy(message, GRN_CTX_MSGSIZE,
-					ctx->errbuf, GRN_CTX_MSGSIZE);
-
+	grn_expr_parse(ctx,
+				   data->expression,
+				   query,
+				   querySize,
+				   data->targetColumn,
+				   GRN_OP_MATCH,
+				   GRN_OP_AND,
+				   data->exprFlags);
+	if (ctx->rc != GRN_SUCCESS)
 		data->expressionHash = 0;
-
-		ereport(ERROR,
-				(errcode(PGrnRCToPgErrorCode(rc)),
-				 errmsg("pgroonga: failed to parse expression: %s: <%.*s>",
-						message,
-						(int) querySize, query)));
-	}
+	PGrnCheck("%s failed to parse expression: <%.*s>",
+			  tag,
+			  (int) querySize, query);
 }
 
 void
 PGrnSequentialSearchDataSetScript(PGrnSequentialSearchData *data,
 								  const char *script,
-								  unsigned int scriptSize)
+								  unsigned int scriptSize,
+								  const char *indexName,
+								  unsigned int indexNameSize)
 {
+	const char *tag = "[sequential-search][query]";
 	grn_expr_flags flags = GRN_EXPR_SYNTAX_SCRIPT;
-	grn_rc rc;
 
 	if (PGrnSequentialSearchDataPrepareExpression(data,
 												  script,
 												  scriptSize,
+												  indexName,
+												  indexNameSize,
 												  PGRN_SEQUENTIAL_SEARCH_SCRIPT))
 	{
 		return;
 	}
 
-	rc = grn_expr_parse(ctx,
-						data->expression,
-						script, scriptSize,
-						data->indexColumnSource,
-						GRN_OP_MATCH, GRN_OP_AND,
-						flags);
-	if (rc != GRN_SUCCESS)
-	{
-		char message[GRN_CTX_MSGSIZE];
-		grn_strncpy(message, GRN_CTX_MSGSIZE,
-					ctx->errbuf, GRN_CTX_MSGSIZE);
-
+	grn_expr_parse(ctx,
+				   data->expression,
+				   script, scriptSize,
+				   data->targetColumn,
+				   GRN_OP_MATCH, GRN_OP_AND,
+				   flags);
+	if (ctx->rc != GRN_SUCCESS)
 		data->expressionHash = 0;
-
-		ereport(ERROR,
-				(errcode(PGrnRCToPgErrorCode(rc)),
-				 errmsg("pgroonga: failed to parse expression: %s: <%.*s>",
-						message,
-						(int) scriptSize, script)));
-	}
+	PGrnCheck("%s failed to parse expression: <%.*s>",
+			  tag,
+			  (int) scriptSize, script);
 }
 
 bool
